//#*#*#*#*#*#*#*#*#*#*#*#*#*#*#*#*#*#*#* DtmCmd.cpp *#*#*#*#*#*#*#*#* (C) 2012-2018 DekTec
//
// DtmCmd - Command line tool for controlling DTM-32xx devices
//
// Please refer to DtmCmd.h for a short description of DtmCmd.

//+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+ LICENSE +=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=

// Copyright (C) 2012-2019 DekTec Digital Video B.V.
//
// Redistribution and use in source and binary forms, with or without modification, are
// permitted provided that the following conditions are met:
//  1. Redistributions of source code must retain the above copyright notice, this list
//     of conditions and the following disclaimer.
//  2. Redistributions in binary format must reproduce the above copyright notice, this
//     list of conditions and the following disclaimer in the documentation.
//  3. The source code may not be modified for the express purpose of enabling hardware
//     features for which no genuine license has been obtained.
//
// THIS SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED,
// INCLUDING BUT NOT LIMITED TO WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
// PURPOSE AND NONINFRINGEMENT.
// IN NO EVENT SHALL DEKTEC DIGITAL VIDEO BV, ITS AGENTS OR ITS EMPLOYEES BE LIABLE FOR
// ANY DIRECT, INDIRECT, CONSEQUENTIAL, INCIDENTAL, OR OTHER DAMAGES (INCLUDING DAMAGES
// FOR THE LOSS OF USE, INFORMATION, GOODWILL, PROFIT, WORK STOPPAGE, DATA, BUSINESS OR
// REVENUE) UNDER ANY CIRCUMSTANCES, OR UNDER ANY LEGAL THEORY, WHETHER IN CONTRACT, IN
// TORT, IN NEGLIGENCE, OR OTHERWISE, ARISING FROM THE USE OF, OR INABILITY TO USE THIS
// SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGES.

//.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.- Includes files -.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.

#include "DtmCmd.h"                 // Command-line tool for controlling DTM-32xx devices

//-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.- Globals -.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.

DtmCmd  DtmCmdApp;                  // The one and only DtmCmd object
Options  g_Opts;                    // Decoded command line options

// Table defining the command-line options of DtmCmd
static DtOptItem  g_CmdOptions[] =
{
    DtOptItem(L"r", g_Opts.m_Read, false, L"Read command"),
    DtOptItem(L"w", g_Opts.m_Write, false, L"Write command"),
    DtOptItem(L"i", g_Opts.m_Index, 0, L"Index number", 0, INT_MAX),
    DtOptItem(L"interface", g_Opts.m_Interface, L"Type interface\n"
              L"  Serial     Serial interface (e.g. COM1)\n"
              L"  I2C        I2C interface (e.g. Nano Rivert Tech Miniboard)"),
    DtOptItem(L"serial", g_Opts.m_Serial,
                    L"Path of the serial port (linux) or name of the com port (windows)"),
    DtOptItem(L"baudrate", g_Opts.m_BaudRate, 0,
                             L"Baudrate of the serial port (e.g. 9600)", 0, MAX_BAUDRATE),
    DtOptItem(L"addr", g_Opts.m_DtmAddress, 0, L"DTM address (e.g. 0x40)",
                                                                       0, MAX_DTMADDRESS),
    DtOptItem(L"upgrade", g_Opts.m_FirmwareUpgrade, L"Firmware upgrade file"),
    DtOptItem(L"usb", g_Opts.m_UsbIf, false, L"Hint that the USB interface is used. This "
                                             L"parameter is required\n for upgrading the "
                                             L"firmware of a DTM-3237 when using the USB "
                                             L"interface"),
    DtOptItem(L"v", g_Opts.m_Verbose, false, L"Verbose mode"),
    DtOptItem(L"?", g_Opts.m_ShowHelp, false, L"Display this help")
};
const int  NUM_CMD_OPTIONS = sizeof(g_CmdOptions) / sizeof(g_CmdOptions[0]);

//-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.- main -.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-
//
int  main(int Argc, char* Argv[])
{
    // Immediately call Main in the context of the DtmCmdApp object
    return DtmCmdApp.Main(Argc, Argv);
}
int  DtmCmd::Main(int Argc, char* Argv[])
{
    try {
        // Parse command line options
        ParseCommandLine(Argc, Argv);
        
        // Show help
        if (g_Opts.m_ShowHelp.ToBool())
        {
            ShowHelp();
            return 0;
        }

        // If the interface option is specified on the command line, store interface
        // related settings in DtmCmd.ini file 
        if (g_Opts.m_Interface.IsSet())
            StoreInterfaceSettings();

        // Load interface settings from DtmCmd.ini file
        LoadDtmCmdIni();

        // Initialize interface
        if (m_CmdLine.m_Interface == L"serial")
            InitSerialInterface();
        else if (m_CmdLine.m_Interface == L"i2c")
            InitI2cInterface();

        // Open firmware file
        if (g_Opts.m_FirmwareUpgrade.IsSet())
        {
            string  Filename(m_CmdLine.m_FwUpgrFile.begin(), 
                                                            m_CmdLine.m_FwUpgrFile.end());
            m_pFwUpgrFile = fopen(Filename.c_str(), "rb");
            if (m_pFwUpgrFile == NULL)
                throw Exc("Cannot open firmware file '%s'", Filename.c_str());
        }
    }
    catch (Exc e) 
    {
        Log(e);
        Log("");
        Log("Use -? option for help");
        return -1;
    }
    catch (DtOptException e) 
    {
        Log(e);
        Log("");
        Log("Use -? option for help");
        return -1;
    }
    catch ( ... ) 
    {
        return -1;
    }

    // Start-up message
    if (g_Opts.m_Verbose.IsSet())
        LogF(L"DtmCmd - DTM-32xx Command Utility v%d.%d.%d (c) 2019 DekTec Digital Video "
             L"B.V.\n",DTMCMD_VERSION_MAJOR, DTMCMD_VERSION_MINOR, DTMCMD_VERSION_BUGFIX);
    
    // Display command info to the user
    if (g_Opts.m_Verbose.IsSet())
        DisplayInfo();
    
    // Initialize DTM handler
    DtmHandlerInit(this, m_CmdLine.m_Interface == L"serial" ? UART : I2C,
                              &RxFunction, &TxFunction, &SleepFunction, &PumpGuiFunction);
    
    // Send DTM command
    SendDtmCommand();
    
    // Close firmware upgrade file
    if (g_Opts.m_FirmwareUpgrade.IsSet())
        fclose(m_pFwUpgrFile);

    // Close interface
    if (m_CmdLine.m_Interface == L"serial")
        CloseSerialInterface();
    
    return 0;
}

//=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+ DtmCmd Main Functions +=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+

//-.-.-.-.-.-.-.-.-.-.-.-.-.-.- DtmCmd::CreateDtmCmdIniFile -.-.-.-.-.-.-.-.-.-.-.-.-.-.-.
//
// Create file 'DtmCmd.ini'. The configuration settings will be read from this file.
//
void  DtmCmd::CreateDtmCmdIniFile(wstring Settings)
{
    // Convert wstring to char
    char  SettingsChar[DTMCMDINI_BUFFER];
    int  WcsLen = (int)wcslen(Settings.c_str());
    if (sizeof(SettingsChar) < WcsLen+1)
        throw Exc("Settings buffer too small");
    wcstombs(SettingsChar, Settings.c_str(), sizeof(SettingsChar));

    // Create a new DtmCmd.ini. If DtmCmd.ini already exists, its contents are discarded 
    // and the file is treated as a new empty file.
    FILE*  pDtmCmdIni = fopen(DTMCMD_CONFIG_FILE, "w+");
    if (pDtmCmdIni == NULL)
        throw Exc("Cannot open file 'DtmCmd.ini' for write");
    int  Size = fwrite(SettingsChar, sizeof(char), strlen(SettingsChar), pDtmCmdIni);
    if (Size == 0)
        throw Exc("Cannot write to file 'DtmCmd.ini'");
    fclose(pDtmCmdIni);
}

//-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.- DtmCmd::DisplayInfo -.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.
//
// Print info about command and configuration
//
void  DtmCmd::DisplayInfo()
{
    if (g_Opts.m_FirmwareUpgrade.IsSet())
        LogF(L"- Firmware upgrade : %ls", m_CmdLine.m_FwUpgrFile.c_str());
    else {
        const char*  CategoryName = g_DtmCategoryInfo[m_CmdLine.m_CategoryIdx].m_pName;
        wstring  wCategoryName = wstring(CategoryName, CategoryName+strlen(CategoryName));
        const char*  SettingName = g_DtmSettingInfo[m_CmdLine.m_SettingsIdx].m_pName;
        wstring  wSettingName = wstring(SettingName, SettingName+strlen(SettingName));

        LogF(L"- Category         : 0x%02X (%ls)",
                                             m_CmdLine.m_Category, wCategoryName.c_str());
        LogF(L"- Setting          : 0x%02X (%ls)",
                                               m_CmdLine.m_Setting, wSettingName.c_str());
        LogF(L"- Index            : 0x%02X", m_CmdLine.m_Index);
    }
    LogF(L"- Interface        : %ls", m_CmdLine.m_Interface.c_str());
    LogF(L"- DTM address      : 0x%02X", wtoi(m_CmdLine.m_DtmAddress));
    if (m_CmdLine.m_Interface == L"serial")
    {
        LogF(L"- Serial path      : %ls", m_CmdLine.m_SerialPath.c_str());
        LogF(L"- Serial baud      : %ls", m_CmdLine.m_SerialBaud.c_str());
    }
    Log("");
}

//-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.- DtmCmd::InitI2cInterface -.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-
//
void  DtmCmd::InitI2cInterface()
{
#if defined(WIN32) || defined(WIN64)
    // Open device
    if (!m_I2cMiniBoard.OpenDevice())
        throw Exc("Cannot open MiniBoard device");

    // Set I2C frequency to 400kbit/s
    NANO_RESULT  Result = m_I2cMiniBoard.Nano_IICSetFrequency(
                                            m_I2cMiniBoard.mb_hDevice, NANO_IIC_FREQ_STD);
    if (Result != NANO_SUCCESS)
        throw Exc("Cannot set frequency");
#else
    // Linux: I2C is not supported
#endif
}

//-.-.-.-.-.-.-.-.-.-.-.-.-.-.- DtmCmd::InitSerialInterface -.-.-.-.-.-.-.-.-.-.-.-.-.-.-.
//
void  DtmCmd::InitSerialInterface()
{
    
    // Get baudrate from command line and check it
    int  BaudRate = wtoi(m_CmdLine.m_SerialBaud);
    if (BaudRate!=9600 && BaudRate!=115200 && BaudRate!=256000)
        throw Exc("Invalid serial baudrate");

#if defined(WIN32) || defined(WIN64)
    //-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.- Windows -.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.

    // The "\\.\" prefix will access the Win32 device namespace instead of the Win32 file 
    // namespace. 
    const std::wstring NtPrefix=L"\\\\.\\";

    // Concatenate NtPrefix with ComPort 
    m_CmdLine.m_SerialPath=NtPrefix+m_CmdLine.m_SerialPath;

    // Open COM port
    m_hSerialPort = CreateFile(m_CmdLine.m_SerialPath.c_str(), 
             GENERIC_READ | GENERIC_WRITE, 0, 0, OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL, 0);

    // Error opening COM port
    if (m_hSerialPort == INVALID_HANDLE_VALUE)
        throw Exc("Cannot open serial port");

    // Get current serial parameters from COM port
    DCB  SerialParams = {0};
    if (!GetCommState(m_hSerialPort, &SerialParams))
        throw Exc("Cannot retrieve serial port settings");

    // Set serial parameters
    switch (BaudRate)
    {
    case 9600: SerialParams.BaudRate = CBR_9600; break;
    case 115200: SerialParams.BaudRate = CBR_115200; break;
    case 256000: SerialParams.BaudRate = CBR_256000; break;
    default: SerialParams.BaudRate = -1; break;
    }
    SerialParams.ByteSize = 8;
    SerialParams.StopBits = ONESTOPBIT;
    SerialParams.Parity = NOPARITY;
    // Flow control can be enabled for reliable RS232 communication during firmware upgrade process.
    SerialParams.fOutxCtsFlow = FALSE;
    SerialParams.fRtsControl = RTS_CONTROL_HANDSHAKE;   // Enable RTS handshaking

    // Write serial parameters to COM port
    if (!SetCommState(m_hSerialPort, &SerialParams))
        throw Exc("Cannot set serial parameters");

    // Get time-out parameters for the COM port
    COMMTIMEOUTS  SerialTimeout = {0};
    if (!GetCommTimeouts(m_hSerialPort, &SerialTimeout))
        throw Exc("Cannot get time-out parameters of serial port");

    // Set time-out parameters
    SerialTimeout.ReadIntervalTimeout = 300;
    SerialTimeout.ReadTotalTimeoutMultiplier = 0;
    SerialTimeout.ReadTotalTimeoutConstant = 1;
    SerialTimeout.WriteTotalTimeoutConstant = 1000;
    SerialTimeout.WriteTotalTimeoutMultiplier = 10;

    // Write time-out parameters to COM port
    if (!SetCommTimeouts(m_hSerialPort, &SerialTimeout))
        throw Exc("Cannot set serial port timing");
#else
    //.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.- Linux -.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-

    // Convert serial path to string
    string  SerialPath(m_CmdLine.m_SerialPath.begin(), m_CmdLine.m_SerialPath.end());
    struct termios  PortSettings;
    
    // Open serial port
    m_SerialPort = open(SerialPath.c_str(), O_RDWR | O_SYNC);
    if (m_SerialPort < 0)
        throw Exc("Cannot open serial port");

    // Get current serial parameters from COM port
    if (tcgetattr(m_SerialPort, &PortSettings))
        throw Exc("Cannot retrieve serial port settings");
    
    // Set serial parameters
    switch (BaudRate)
    {
    case 9600: 
        cfsetospeed(&PortSettings, B9600); 
        cfsetispeed(&PortSettings, B9600); 
        break;
    case 115200:
        cfsetospeed(&PortSettings, B115200);
        cfsetispeed(&PortSettings, B115200);
        break;
    case 256000:
#if defined(B256000)   
        cfsetospeed(&PortSettings, B256000);
        cfsetispeed(&PortSettings, B256000);
        break;
#else
        throw Exc("Cannot set serial parameters: B256000 is not supported by termios.h");
#endif
    default:
        throw Exc("Cannot set serial parameters: Invalid baudrate specified");
    }
    
    PortSettings.c_iflag |= IGNBRK | IGNCR;
    PortSettings.c_iflag &= ~(BRKINT | PARMRK | ISTRIP | INLCR | ICRNL | IXON | IXANY 
                                                                 | INPCK | IUCLC | IXOFF);
    PortSettings.c_oflag &= ~(OPOST | OLCUC | ONLCR | OCRNL | ONLRET | OFILL | NLDLY 
                                        | CRDLY | TABDLY | BSDLY | VTDLY | FFDLY | ONOCR);
    PortSettings.c_cflag &= ~(CSIZE | PARENB | CSTOPB | CMSPAR);
    PortSettings.c_cflag |= (CS8 | CLOCAL | CREAD | CRTSCTS | HUPCL);
    PortSettings.c_lflag &= ~(ECHO | ECHONL | ECHOE | ECHOK | ICANON | ISIG | IEXTEN |
                                                          ISIG | ICANON | XCASE | TOSTOP);
    PortSettings.c_cc[VMIN] = 60;
    PortSettings.c_cc[VTIME] = 1;

    if (tcsetattr(m_SerialPort, TCSANOW, &PortSettings))
        throw Exc("Cannot set serial parameters");
#endif
}

//-.-.-.-.-.-.-.-.-.-.-.-.-.-.- DtmCmd::CloseSerialInterface -.-.-.-.-.-.-.-.-.-.-.-.-.-.-
//
void  DtmCmd::CloseSerialInterface()
{
#if defined(WIN32) || defined(WIN64)
    CloseHandle(m_hSerialPort);
    m_hSerialPort = NULL;
#else
    close(m_SerialPort);
    m_SerialPort = -1;
#endif
}

//.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.- DtmCmd::LoadDtmCmdIni -.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-
//
// Load and decode "DtmCmd.ini".
// The decoded configuration settings are stored in m_CmdLine.
//
void  DtmCmd::LoadDtmCmdIni()
{
    // Open DtmCmd.ini
    FILE*  pDtmCmdIni = fopen(DTMCMD_CONFIG_FILE, "r");
    if (pDtmCmdIni == NULL)
        throw Exc("Could not find/read DtmCmd.ini file");

    // Step through each line of the configuration file 
    wchar_t  Buffer[DTMCMDINI_BUFFER];
    while (fgetws(Buffer, DTMCMDINI_BUFFER, pDtmCmdIni) != NULL)
    {
        // Convert line buffer to std::wstring and strip newline at the end, if any
        wstring  Line(Buffer);
        if (Line[Line.length()-1] == L'\n')
            Line = Line.substr(0, Line.length()-1);

        // Get value of configuration setting in this line and store in m_CmdLine
        if (Line.find(L"Interface=") == 0)
            m_CmdLine.m_Interface = Line.substr(Line.find(L"=")+1, Line.length()-1);
        else if (Line.find(L"DtmAddress=") == 0)
            m_CmdLine.m_DtmAddress = Line.substr(Line.find(L"=")+1, Line.length()-1);
        else if (Line.find(L"SerialPath=") == 0)
            m_CmdLine.m_SerialPath = Line.substr(Line.find(L"=")+1, Line.length()-1);
        else if (Line.find(L"BaudRate=") == 0)
            m_CmdLine.m_SerialBaud = Line.substr(Line.find(L"=")+1, Line.length()-1);
    }
    fclose(pDtmCmdIni);

    // Validate settings
    if (m_CmdLine.m_Interface!=L"serial" && m_CmdLine.m_Interface!=L"i2c")
        throw Exc("Invalid interface in DtmCmd.ini file");
    if (m_CmdLine.m_DtmAddress.length() == 0)
        throw Exc("DTM address missing in DtmCmd.ini file");
    if (m_CmdLine.m_Interface==L"serial" && m_CmdLine.m_SerialPath.length()==0)
        throw Exc("Serial path missing in DtmCmd.ini file");
    if (m_CmdLine.m_Interface==L"serial" && m_CmdLine.m_SerialBaud.length()==0)
        throw Exc("Serial baudrate missing in DtmCmd.ini file");
    if (m_CmdLine.m_Interface == L"serial")
    {
        int  BaudRate = wtoi(m_CmdLine.m_SerialBaud);
        if (BaudRate<0 || BaudRate>MAX_BAUDRATE)
            throw Exc("Invalid baudrate for serial port");
    }
    int  DtmAddress = wtoi(m_CmdLine.m_DtmAddress);
    if (DtmAddress<0 || DtmAddress>MAX_DTMADDRESS)
        throw Exc("Invalid DTM address");
}

//-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.- DtmCmd::ParseCommandLine -.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-
//
void  DtmCmd::ParseCommandLine(int argc, char* argv[])
{
    // Parse command line options into global structure g_Opts
    list<wstring>  FreeArgs;
    DtOptItem::ParseOpt(g_CmdOptions, argc, argv, FreeArgs);

    // Don't parse any further if help option is specified
    if (g_Opts.m_ShowHelp.IsSet())
        return;

    // If firmware upgrade option is set, get firmware filename and return
    if (g_Opts.m_FirmwareUpgrade.IsSet())
    {
        m_CmdLine.m_FwUpgrFile = g_Opts.m_FirmwareUpgrade.ToString();
        return;
    }

    // Check for invalid parameter options and required parameters

    // If one of the interface configuration options is set, all configuration options
    // relevant to the interface must be set
    if (g_Opts.m_Interface.IsSet() || g_Opts.m_Serial.IsSet()
                              || g_Opts.m_BaudRate.IsSet() || g_Opts.m_DtmAddress.IsSet())
    {
        if (g_Opts.m_Interface.ToLower() != L"serial"
                                                && g_Opts.m_Interface.ToLower() != L"i2c")
            throw Exc("Invalid interface");
        if (!g_Opts.m_DtmAddress.IsSet())
            throw Exc("Missing DTM address");
        if (g_Opts.m_Interface.ToLower() == L"serial")
            if (!g_Opts.m_Serial.IsSet() || !g_Opts.m_BaudRate.IsSet())
                throw Exc("Missing serial or baudrate argument");
    }
    if (FreeArgs.size() == 0)
        throw Exc("Missing category and setting argument");
    if (FreeArgs.size() == 1)
        throw Exc("Missing setting argument");
    if (g_Opts.m_Write.IsSet() && FreeArgs.size()<3)
        throw Exc("Missing data argument");
    if (!g_Opts.m_Read.IsSet() && !g_Opts.m_Write.IsSet())
        throw Exc("No read or write command has been specified");
    if (g_Opts.m_Read.IsSet() && g_Opts.m_Write.IsSet())
        throw Exc("Read and write commands may not be specified simultaneously");

    // Parse parameters of the DTM-32xx setting: category, setting number and index
    std::list<wstring>::iterator  FreeArgsIt = FreeArgs.begin();
    m_CmdLine.m_Category = wtoi((FreeArgsIt++)->c_str());
    m_CmdLine.m_Setting = wtoi((FreeArgsIt++)->c_str());
    m_CmdLine.m_Index = g_Opts.m_Index.ToInt();
    
    // Find category in g_DtmCategoryInfo (this is a table defined in DtmHandler.c with
    // information about each DTM-32xx settings category)
    m_CmdLine.m_CategoryIdx = -1;
    for (int Idx=0; Idx<g_NumDtmCategoryInfoItems; Idx++)
    {
        if (g_DtmCategoryInfo[Idx].m_Category == m_CmdLine.m_Category)
        {
            m_CmdLine.m_CategoryIdx = Idx;
            break;
        }
    }
    if (m_CmdLine.m_CategoryIdx < 0)
        throw Exc("Unknown category");

    // Find setting in g_DtmSettingInfo (table with info about each DTM-32xx setting)
    m_CmdLine.m_SettingsIdx = -1;
    for (int Idx=0; Idx<g_NumDtmSettingInfoItems; Idx++)
    {
        if (g_DtmSettingInfo[Idx].m_Category==m_CmdLine.m_Category 
                                  && g_DtmSettingInfo[Idx].m_Setting==m_CmdLine.m_Setting)
        {
            m_CmdLine.m_SettingsIdx = Idx;
            break;
        }
    }
    if (m_CmdLine.m_SettingsIdx < 0)
        throw Exc("Unknown setting");

    // For convenience, save DTM-32xx data type and units in m_CmdLine
    m_CmdLine.m_DataType = g_DtmSettingInfo[m_CmdLine.m_SettingsIdx].m_DataType;
    m_CmdLine.m_Units = g_DtmSettingInfo[m_CmdLine.m_SettingsIdx].m_Units;

    // For write commands, get data to be written
    int  NumFieldsConv;
    if (g_Opts.m_Write.IsSet())
    {
        switch (m_CmdLine.m_DataType)
        {
        case D_UINT8: 
        case D_UINT16: 
        case D_INT32: 
        case D_UINT32: 
        case D_UINT64: 
            NumFieldsConv = swscanf(FreeArgsIt->c_str(), L"0x%llx", &m_CmdLine.m_Data);
            if (NumFieldsConv != 1)
                NumFieldsConv = swscanf(FreeArgsIt->c_str(), L"%lld", &m_CmdLine.m_Data);
            if (NumFieldsConv != 1)
                throw Exc("Invalid data format");
            break;

        case D_IP:
            {
                int  Ip[4];
                NumFieldsConv = swscanf(FreeArgsIt->c_str(), L"%u.%u.%u.%u", 
                                                          &Ip[0], &Ip[1], &Ip[2], &Ip[3]);
                if (NumFieldsConv != 4)
                    throw Exc("Invalid IP address");
                for (int i=0; i<4; i++)
                    if (Ip[i] > 255)
                        throw Exc("Invalid IP address");
                m_CmdLine.m_Data = ((long long)Ip[0]<<24)
                                                       | (Ip[1]<<16) | (Ip[2]<<8) | Ip[3];
            }
            break;

        case D_IPV6:
            {
                std::wstring  StringIp(*FreeArgsIt);
                std::vector<std::wstring>  Bits;
                std::wstring::size_type  i = 0;
                // Find all pieces seperated by :'s
                while (true)
                {
                    std::wstring::size_type  Seperator = StringIp.find(L':',i);
                    if (Seperator == std::wstring::npos)
                    {
                        Bits.push_back(StringIp.substr(i));
                        break;
                    } else {
                        Bits.push_back(StringIp.substr(i, Seperator - i));
                        i = Seperator + 1;
                        continue;
                    }
                }
                // If less than 3 groups, only one colon was encountered,
                // which is always an invalid address
                if (Bits.size() < 3)
                    throw Exc("Not enough groups were specified for an IPv6 address");
                // If the first group is empty (e.g. starts with :)
                // then the only valid IPv6 address starts with ::...
                // In that case, leave just one of the empty groups to expand later
                if (Bits.front().length() == 0)
                {
                    Bits.erase(Bits.begin());
                    if (Bits.front().length() != 0)
                        throw Exc("IPv6 address can not start with a single ':'");
                }
                // Same for the last group (ends with :)
                // In case of ::, both ifs would each remove one empty group, and leave
                // a singular empty group to be expanded to all zeroes
                if (Bits.back().length() == 0)
                {
                    Bits.pop_back();
                    if (Bits.back().length() != 0)
                        throw Exc("IPv6 address can not end with a single ':'");
                }
                if (Bits.size() > 8)
                    throw Exc("IPv6 address consists of too many groups");
                //Possibly expand single ::
                for (std::vector<std::wstring>::iterator  i = Bits.begin();
                                                                 i != Bits.end(); i++)
                {
                    if (i->length() == 0) {
                        i = Bits.erase(i);
                        Bits.insert(i, 8 - Bits.size(), L"0");
                        break;
                    }
                }

                for (std::vector<std::wstring>::iterator  i = Bits.begin();
                                                                     i != Bits.end(); i++)
                {
                    if (i->length() == 0)
                        throw Exc("Only one empty group allowed in an IPv6 address");
                    wchar_t*  End;
                    unsigned long  NumBit = wcstoul(i->c_str(), &End, 16);
                    if (End != &i->c_str()[i->length()])
                        throw Exc("Only hexadecimal characters allowed "
                                                                 "in IPv6 address group");
                    if (NumBit > 0xFFFF)
                        throw Exc("IPv6 address group out of range (>0xFFFF)");
                    size_t  index = i - Bits.begin();
                    m_CmdLine.m_Bytes[(index*2)+0] = ((NumBit >> 8)&0xFF);
                    m_CmdLine.m_Bytes[(index*2)+1] = ((NumBit >> 0)&0xFF);
                }
            }
            break;

        case D_MAC:
            {
                int  Mac[6];
                NumFieldsConv = swscanf(FreeArgsIt->c_str(), L"%x-%x-%x-%x-%x-%x", 
                                    &Mac[0], &Mac[1], &Mac[2], &Mac[3], &Mac[4], &Mac[5]);
                if (NumFieldsConv != 6)
                    throw Exc("Invalid MAC address");
                for (int i=0; i<6; i++)
                    if (Mac[i] > 255)
                        throw Exc("Invalid MAC address");
                m_CmdLine.m_Data = ((long long)Mac[0]<<40) |  ((long long)Mac[1]<<32) 
                         |  ((long long)Mac[2]<<24) | (Mac[3]<<16) | (Mac[4]<<8) | Mac[5];
            }
            break;

        default:
            throw Exc("Invalid data type");
            break;
        }
    }
}

//.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.- DtmCmd::PrintfDtmError -.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.
//
void  DtmCmd::PrintfDtmError(DtmResult Result)
{
    wstring  DtmError;
    switch (Result)
    {
    case DTM_OK: DtmError = L"DTM_OK"; break;
    case DTM_E_ERROR_REPLY: DtmError = L"DTM_E_ERROR_REPLY"; break;
    case DTM_E_INCOMPLETE_MSG: DtmError = L"DTM_E_INCOMPLETE_MSG"; break;
    case DTM_E_INCORRECT_CHECKSUM: DtmError = L"DTM_E_INCORRECT_CHECKSUM"; break;
    case DTM_E_INCORRECT_REPLY: DtmError = L"DTM_E_INCORRECT_REPLY"; break;
    case DTM_E_INVALID_PARAMETER: DtmError = L"DTM_E_INVALID_PARAMETER"; break;
    case DTM_E_INVALID_MODE: DtmError = L"DTM_E_INVALID_MODE"; break;
    case DTM_E_NULL_PNTR: DtmError = L"DTM_E_NULL_PNTR"; break;
    case DTM_E_OUT_OF_MEM: DtmError = L"DTM_E_OUT_OF_MEM"; break;
    case DTM_E_PORT_CLOSED: DtmError = L"DTM_E_PORT_CLOSED"; break;
    case DTM_E_RX_FAILED: DtmError = L"DTM_E_RX_FAILED"; break;
    case DTM_E_RXFUNC_NULL_PNTR: DtmError = L"DTM_E_RXFUNC_NULL_PNTR"; break;
    case DTM_E_TIME_OUT: DtmError = L"DTM_E_TIME_OUT"; break;
    case DTM_E_TX_FAILED: DtmError = L"DTM_E_TX_FAILED"; break;
    case DTM_E_TXFUNC_NULL_PNTR: DtmError = L"DTM_E_TXFUNC_NULL_PNTR"; break;
    case DTM_E_UPG_FAILED: DtmError = L"DTM_E_UPG_FAILED"; break;
    case DTM_E_UPG_FILE_ERROR: DtmError = L"DTM_E_UPG_FILE_ERROR"; break;
    case DTM_E_UPG_FILE_INVALID_CHECKSUM: DtmError = L"DTM_E_UPG_FILE_INVALID_CHECKSUM"; 
                                                                                    break;
    case DTM_E_UPG_VERIFY_FAILED: DtmError = L"DTM_E_UPG_VERIFY_FAILED"; break;
    case DTM_E_BUSY_REPLY: DtmError = L"DTM_E_BUSY_REPLY"; break;
    case DTM_E_UPG_NULL_PNTR: DtmError = L"DTM_E_UPG_NULL_PNTR"; break;
    case DTM_E_NOT_SUPPORTED: DtmError = L"DTM_E_NOT_SUPPORTED"; break;
    case DTM_GOT_LESS_DATA: DtmError = L"DTM_GOT_LESS_DATA"; break;
    case DTM_GOT_MORE_DATA: DtmError = L"DTM_GOT_MORE_DATA"; break;
    case DTM_GOT_NO_DATA: DtmError = L"DTM_GOT_NO_DATA"; break;
    case DTM_REPLY_IGNORED: DtmError = L"DTM_REPLY_IGNORED"; break;
    default: DtmError = L"UNKOWN"; break;
    }

    wstring  Error = L"An error occurred during communication with the DTM-32xx device.\n"
                     L"DtmHandler error code: ";
    Error += DtmError;
    Log(Error.c_str());
}

// Helper macro: If DtmHander result is not OK, print error and return
#define CHECK_RESULT(Result)     if (Result != DTM_OK) { PrintfDtmError(Result); return; }

//.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.- DtmCmd::SendDtmCommand -.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.
//
// This function sends a read, write or firmware upgrade command to the DTM-32xx module
// by using the corresponding DtmHandler API functions
//
void  DtmCmd::SendDtmCommand()
{
    int  Data;                      // Helper integer data variable
    unsigned long long  D64;        // Helper 64-bit integer data variable
    DtmResult  Result;              // Result of DtmHandler API function

    int  DtmAddress = wtoi(m_CmdLine.m_DtmAddress);
    int  Category = (int)(m_CmdLine.m_Category);
    int  Setting = (int)(m_CmdLine.m_Setting);
    int  Index = (int)(m_CmdLine.m_Index);

    if (g_Opts.m_FirmwareUpgrade.IsSet())
    {
        // Firmware upgrade
        int  FileSize, UpgradeType, Mode;
        
        // Switch from normal operation to firmware upgrade mode for DTM type 2 devices
        Result = DtmGetUpgradeType(DtmAddress, &UpgradeType);
        CHECK_RESULT(Result);
        if (UpgradeType == 2)
        {
            // Read current mode
            Result = DtmReadInt(DtmAddress, 0x85, 0x01, 0x00, &Mode); 
            CHECK_RESULT(Result);
            if (Mode != 1)
            {
                // Switch to firmware upgrade mode
                Result = DtmWriteInt(DtmAddress, 0x85, 0x01, 0x00, 1);
                CHECK_RESULT(Result);

                // DTM type 2 devices disable their USB interface when switching mode
                // Close handle to serial port if USB interface is used
                if (g_Opts.m_UsbIf.IsSet())
                    CloseSerialInterface();

                // Wait for the DTM to initialize firmware upgrade mode
#ifdef WINBUILD
                Sleep(3000);
#else
                usleep(3000000);
#endif
                // Initialize serial interface if USB interface is used
                if (g_Opts.m_UsbIf.IsSet())
                    InitSerialInterface();
            }
        }

        // Get filesize
        fseek(m_pFwUpgrFile, 0, SEEK_END);
        FileSize = ftell(m_pFwUpgrFile);
        fseek(m_pFwUpgrFile, 0, SEEK_SET);

        // Upgrade firmware
        Result = DtmFirmwareUpgrade(DtmAddress, FileSize, &UpgGetData, 
                                                                      &UpgReportProgress);
        if (g_Opts.m_Verbose.IsSet())
            LogF(L"");
        CHECK_RESULT(Result);

        // Switch to normal operation for DTM type 2 devices
        if (UpgradeType == 2)
        {
            // Switch to normal operation
            Result = DtmWriteInt(DtmAddress, 0x85, 0x01, 0x00, 0);
            CHECK_RESULT(Result);

            // DTM type 2 devices disable their USB interface when switching mode
            // Close handle to serial port if USB interface is used
            if (g_Opts.m_UsbIf.IsSet())
                CloseSerialInterface();

            // Wait for the DTM to initialize normal operation mode
#ifdef WINBUILD
                Sleep(10000);
#else
                usleep(10000000);
#endif
            // Initialize serial interface if USB interface is used
            if (g_Opts.m_UsbIf.IsSet())
                InitSerialInterface();
        }
    }
    else if (g_Opts.m_Write.IsSet())
    {
        // Write command
        int  Data;
        switch (m_CmdLine.m_DataType)
        {
        case D_UINT8:
        case D_UINT16:
        case D_INT32:
        case D_UINT32:
        case D_IP:
            Data = (int)(m_CmdLine.m_Data);
            Result = DtmWriteInt(DtmAddress, Category, Setting, Index, Data);
            CHECK_RESULT(Result);
            if (g_Opts.m_Verbose.IsSet())
                LogF(L"- Data written     : %u (0x%X)", Data, Data);
            break;

        case D_IPV6:
            Result = DtmWriteBytes(DtmAddress, Category, Setting, Index,
                                                           (char *)m_CmdLine.m_Bytes, 16);
            CHECK_RESULT(Result);
            if (g_Opts.m_Verbose.IsSet())
                LogF(L"- Data written     : "
                                            L"%02X %02X %02X %02X %02X %02X %02X %02X "
                                            L"%02X %02X %02X %02X %02X %02X %02X %02X",
                                            m_CmdLine.m_Bytes[0], m_CmdLine.m_Bytes[1],
                                            m_CmdLine.m_Bytes[2], m_CmdLine.m_Bytes[3],
                                            m_CmdLine.m_Bytes[4], m_CmdLine.m_Bytes[5],
                                            m_CmdLine.m_Bytes[6], m_CmdLine.m_Bytes[7],
                                            m_CmdLine.m_Bytes[8], m_CmdLine.m_Bytes[9],
                                            m_CmdLine.m_Bytes[10], m_CmdLine.m_Bytes[11],
                                            m_CmdLine.m_Bytes[12], m_CmdLine.m_Bytes[13],
                                            m_CmdLine.m_Bytes[14], m_CmdLine.m_Bytes[15]);
            break;

        case D_UINT64:
        case D_MAC:
            Result = DtmWriteInt64(DtmAddress, Category, Setting, Index, 
                                                  (unsigned long long)(m_CmdLine.m_Data));
            CHECK_RESULT(Result);
            if (g_Opts.m_Verbose.IsSet())
                LogF(L"- Data written     : %llu (0x%X)", m_CmdLine.m_Data, 
                                                                        m_CmdLine.m_Data);
            break;
        }
    }
    else if (g_Opts.m_Read.IsSet())
    {
        // Read command
        switch (m_CmdLine.m_DataType)
        {
        case D_UINT8:
        case D_UINT16:
        case D_INT32:
        case D_UINT32:
        case D_IP:
            Result = DtmReadInt(DtmAddress, Category, Setting, Index, &Data);
            CHECK_RESULT(Result);
            if (m_CmdLine.m_DataType == D_INT32)
                LogF(L"- Data read        : %.2f", float(Data) / m_CmdLine.m_Units);
            else if (m_CmdLine.m_DataType == D_IP)
                LogF(L"- Data read        : %u.%u.%u.%u", ((Data>>24)&0xFF), 
                                        ((Data>>16)&0xFF), ((Data>>8)&0xFF), (Data&0xFF));
            else
                LogF(L"- Data read        : %u (0x%X)", Data, Data);
            break;

        case D_IPV6:
        {
            unsigned char  IpAddress[16];
            Result = DtmReadBytes(DtmAddress, Category, Setting, Index,
                                                                   (char*)&IpAddress, 16);
            LogF(L"- Data read        : %X:%X:%X:%X:%X:%X:%X:%X",
              (IpAddress[0] << 8) | IpAddress[1], (IpAddress[2] << 8) | IpAddress[3],
              (IpAddress[4] << 8) | IpAddress[5], (IpAddress[6] << 8) | IpAddress[7],
              (IpAddress[8] << 8) | IpAddress[9], (IpAddress[10] << 8) | IpAddress[11],
              (IpAddress[12] << 8) | IpAddress[13], (IpAddress[14] << 8) | IpAddress[15]);
            break;
        }

        case D_UINT64:
        case D_MAC:
            Result = DtmReadInt64(DtmAddress, Category, Setting, Index, &D64);
            CHECK_RESULT(Result);
            if (m_CmdLine.m_DataType == D_UINT64)
            {
                // Print statistics as float
                if (Category == 0x92)
                {
                    double  Statistic = double(D64) / pow(2.0, 64.0);
                    LogF(L"- Data read        : %E", Statistic);
                } else {
                    LogF(L"- Data read        : %llu (0x%llX)", D64, D64);
                }
            }
            else 
                LogF(L"- Data read        : %02X-%02X-%02X-%02X-%02X-%02X", 
                      (int)((D64>>40)&0xFF), (int)((D64>>32)&0xFF), (int)((D64>>24)&0xFF),
                      (int)((D64>>16)&0xFF), (int)((D64>>8)&0xFF), (int)(D64&0xFF));
            break;
        }
    }
}

//-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.- DtmCmd::ShowHelp -.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-
//
void  DtmCmd::ShowHelp()
{
    Log("" );
    Log("Usage:");
    Log( "Usage:");
    Log( "   DtmCmd [-r|-w|-upgrade filename] [-i index] [-interface type] [-usb]\n"
         "          [-serial path_name] [-baudrate baud] [-addr dtmaddress] [-?]\n"
         "          Category Setting [Data]");
    Log("");
    Log("Where:");
    Log("   Category is the category number (decimal or hexadecimal)");
    Log("   Setting is the setting number (decimal or hexadecimal)");
    Log("   [Data] is the data to write (decimal, hexadecimal, IP or MAC address)");
    Log("");
    Log("NOTE: It is recommended to use the -usb parameter when using the USB interface");
    Log("NOTE: It is required to use -- separator when using a negative number as data");
    Log( "");
    Log("Options:");
    DtOptItem::PrintOptionsHelp(g_CmdOptions, NUM_CMD_OPTIONS);
    Log("Examples:");
    Log("   DtmCmd -interface Serial -serial COM4 -baudrate 9600 -addr 0x60 -r 1 5");
    Log("   DtmCmd -interface Serial -serial /dev/ttyACM0 -baudrate 9600 -addr 0x60 \n"
        "          -r 1 5");
    Log("   DtmCmd -interface Serial -serial COM4 -baudrate 9600 -addr 0x60 -w 0x90 1 1");
    Log("   DtmCmd -interface I2C -addr 0x60 -w 0x90 0x03 950000000");
    Log("   DtmCmd -interface Serial -serial COM4 -baudrate 9600 -addr 0x60 \n"
        "          -upgrade Dtm3237FwV0.dtm");
    Log("   DtmCmd -interface Serial -serial COM5 -usb -baudrate 256000 -addr 0x60 \n"
        "          -upgrade Dtm3237FwV0.dtm");
    Log("   DtmCmd -interface I2C -addr 0x60 -upgrade Dtm3237FwV0.dtm");
}

//.-.-.-.-.-.-.-.-.-.-.-.-.-.- DtmCmd::StoreInterfaceSettings -.-.-.-.-.-.-.-.-.-.-.-.-.-.
//
// Save the interface-related settings in DtmCmd.ini.
// For serial interface, both the serial path and baudrate have to be specified.
//
void  DtmCmd::StoreInterfaceSettings()
{
    wstring  Settings;
    Settings = L"Interface=";
    Settings += g_Opts.m_Interface.ToLower();
    Settings += L"\n";
    Settings += L"DtmAddress=";
    Settings += g_Opts.m_DtmAddress.ToLower();
    Settings += L"\n";

    // Serial interface options
    if (g_Opts.m_Interface.ToLower() == L"serial")
    {
        Settings += L"SerialPath=";
        Settings += g_Opts.m_Serial.ToString();
        Settings += L"\n";
        Settings += L"BaudRate=";
        Settings += g_Opts.m_BaudRate.ToLower();
    }

    // Write settings string to DtmCmd.ini
    CreateDtmCmdIniFile(Settings);
}

//=+=+=+=+=+=+=+=+=+=+=+=+=+=+ DTM Handler Callback Functions +=+=+=+=+=+=+=+=+=+=+=+=+=+=

//.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.- RxFunction -.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.
//
// Function called by the DTM handler to receive data from I2C or RS-232/422/485 port
//
DtmResult  RxFunction(void* pOpaque, int DtmAddr, char* pData, int Expected, 
                                                              int* pReturned, int TimeOut)
{
    DtmCmd*  pDtmCmd = (DtmCmd*)pOpaque;         // Pointer to DtmCmd object

#if defined(WIN32) || defined(WIN64)
    // Windows implementation
    if (pDtmCmd->m_CmdLine.m_Interface == L"serial")
    {
        DWORD  TotalNumBytesRead = 0;

        // Time-out implementation (COMMTIMEOUTS doesn't work for USB serial emulator)
        DWORD  StartTime = ::GetTickCount();
        DWORD  CurrentTime = StartTime;
        do {
            DWORD  NumBytesReadThisCall;
            ReadFile(pDtmCmd->m_hSerialPort, &pData[TotalNumBytesRead],
                                 Expected-TotalNumBytesRead, &NumBytesReadThisCall, NULL);
            TotalNumBytesRead += NumBytesReadThisCall;
            if (TotalNumBytesRead == Expected)
                break;
            CurrentTime = ::GetTickCount();
        }
        while ((int)(CurrentTime-StartTime) < TimeOut);

        *pReturned = TotalNumBytesRead;
        if (TotalNumBytesRead == 0) 
            return DTM_E_RX_FAILED; // Didn't receive anything
        if (TotalNumBytesRead != Expected)
            return DTM_E_TIME_OUT;  // Did receive something, but not enough: Timeout
        return DTM_OK;
    }
    else if (pDtmCmd->m_CmdLine.m_Interface == L"i2c")
    {
        NANO_RESULT  Result = pDtmCmd->m_I2cMiniBoard.Nano_IICRead(
                                       pDtmCmd->m_I2cMiniBoard.mb_hDevice, 
                                       (BYTE)DtmAddr, 0, 0, (BYTE)Expected, (BYTE*)pData);
        // #bytres returned is not implemented for I2C
        if (Result == NANO_SUCCESS)
            *pReturned = Expected;  
        else
            *pReturned = -1;
        return (Result != NANO_SUCCESS) ? DTM_E_RX_FAILED : DTM_OK;
    }
    
    return DTM_E_RX_FAILED;
#else
    // Linux implementation
    if (pDtmCmd->m_CmdLine.m_Interface == L"serial")
    {
        int  TotalNumBytesRead = 0;
        struct pollfd  SerialPoll;

        SerialPoll.fd = pDtmCmd->m_SerialPort;
        SerialPoll.events = POLLIN | POLLRDHUP | POLLHUP;

        while (TotalNumBytesRead < Expected)
        {
            if (poll(&SerialPoll, 1, TimeOut) != 1)
            {
                *pReturned = TotalNumBytesRead;
                return DTM_E_TIME_OUT;  // Timeout
            }

            if (SerialPoll.revents & POLLIN)
            {
                TotalNumBytesRead += read(pDtmCmd->m_SerialPort, 
                                   pData + TotalNumBytesRead, Expected-TotalNumBytesRead);
            } else {
                // POLLRDHUP or POLLHUP
                *pReturned = TotalNumBytesRead;
                return DTM_E_PORT_CLOSED; 
            }
        }
        *pReturned = TotalNumBytesRead;
        return DTM_OK;
    }
    else if (pDtmCmd->m_CmdLine.m_Interface == L"i2c")
    {
        // Not supported in DtmCmd for Linux
        return DTM_E_NOT_SUPPORTED;
    }

    return DTM_E_RX_FAILED;
#endif
}

//.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.- SleepFunction -.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-
//
// Function called by the DTM handler to suspend execution for a given period of time
//
DtmResult  SleepFunction(void* pOpaque, int Milliseconds)
{
#if defined(WIN32) || defined(WIN64)
    // Set the minimum supported timer resolution
    TIMECAPS  TimeCaps;
    MMRESULT  Result = timeGetDevCaps(&TimeCaps, sizeof(TIMECAPS));
    if (Result == MMSYSERR_NOERROR)
        Result = timeBeginPeriod(TimeCaps.wPeriodMin);
    
    // Sleep
    Sleep(Milliseconds);
    
    // Clear previously set minimum timer resolution
    if (Result == TIMERR_NOERROR)
        timeEndPeriod(TimeCaps.wPeriodMin);

    return DTM_OK;
#else
    usleep(Milliseconds*1000);
#endif
}

//.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.- TxFunction -.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.
//
// Function called by the DTM handler to receive data from I2C or RS-232/422/485 port
//
DtmResult  TxFunction(void* pOpaque, int DtmAddr, char* pData, int Length)
{
    DtmCmd*  pDtmCmd = (DtmCmd*)pOpaque;         // Pointer to DtmCmd object

#if defined(WIN32) || defined(WIN64)
    // Windows implemantation
    if (pDtmCmd->m_CmdLine.m_Interface == L"serial")
    {
        DWORD  NumBytesWritten;
        if (!WriteFile(pDtmCmd->m_hSerialPort, pData, Length, &NumBytesWritten, NULL))
            return DTM_E_TX_FAILED;
        return (NumBytesWritten != Length) ? DTM_E_TX_FAILED : DTM_OK;
    }
    else if (pDtmCmd->m_CmdLine.m_Interface == L"i2c")
    {
        NANO_RESULT  Result = pDtmCmd->m_I2cMiniBoard.Nano_IICWrite(
                                 pDtmCmd->m_I2cMiniBoard.mb_hDevice, 
                                 (BYTE)(DtmAddr), 0, 0, (BYTE)(Length), (BYTE*)(pData));
        return (Result != NANO_SUCCESS) ? DTM_E_TX_FAILED : DTM_OK;
    }
#else
    // Linux implementation
    if (pDtmCmd->m_CmdLine.m_Interface == L"serial")
    {
        int  NumBytesWritten = write(pDtmCmd->m_SerialPort, pData, Length);
        return (NumBytesWritten != Length) ? DTM_E_TX_FAILED : DTM_OK;
    }
    else if (pDtmCmd->m_CmdLine.m_Interface == L"i2c")
    {
        return DTM_E_NOT_SUPPORTED;
    }
#endif
    return DTM_E_TX_FAILED;
}

//.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.- UpgGetData -.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.
//
// Function called by the DTM handler to read data from a firmware file
//
DtmResult  UpgGetData(void* pOpaque, char* pData, int NumToRead)
{
    size_t  NumBytesRead = fread(pData, sizeof(char), (size_t)(NumToRead), 
                                                                 DtmCmdApp.m_pFwUpgrFile);
    return NumBytesRead == (size_t)(NumToRead) ? DTM_OK : DTM_E_UPG_FILE_ERROR;
}

//.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.- UpgReportProgress -.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-
//
// Function called by the DTM handler to report progress of firmware upgrading
//
void  UpgReportProgress(void* pOpaque, DtmUpgPhase Phase, int Progress)
{
    if (!g_Opts.m_Verbose.IsSet())
        return;

    wstring  CurrentPhase;
    switch (Phase)
    {
    case INIT: CurrentPhase = L"Init"; break;
    case UPLOADING: CurrentPhase = L"Uploading"; break;
    case PROGRAMMING: CurrentPhase = L"Programming"; break;
    case FINISHED: CurrentPhase = L"Finished"; break;
    default: CurrentPhase = L"Unkown"; break;
    }
    
    // Clear line and print updated progress
    wprintf(L"\r                                                                       ");
    wprintf(L"\rCurrent phase: %ls, progress: %d", CurrentPhase.c_str(), Progress);
}

//-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.- PumpGuiFunction -.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.
//
// Function called by the DTM handler to let us process GUI events
//
void  PumpGuiFunction(void* pOpaque)
{
    // We have no GUI, so nothing needs to be done
}

//=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+ Low-Level Helper Functions +=+=+=+=+=+=+=+=+=+=+=+=+=+=+=

//-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.- Log -.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.
//
void  Log(const wchar_t* pMessage)
{
    // Print message and add new line
    wprintf(pMessage); wprintf(L"\n");
}
void  Log(const char* pMessage)
{
    // Under linux we can't mix printf and wprintf, so convert
    // to wstring and always use wprintf
    string  Str(pMessage);
    Log(wstring(Str.begin(), Str.end()).c_str());
}
void  LogF(const wchar_t* pMessage, ... )
{
    wchar_t sz[512];

    va_list  ArgList;
    va_start(ArgList, pMessage);
    _vsnwprintf(sz, (sizeof(sz)/sizeof(wchar_t))-1, pMessage, ArgList);
    va_end(ArgList);

    Log(sz);
}

//-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.- wtoi -.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-
//
int  wtoi(wstring Str)
{
    long  Result;
    if (Str.size()>2 && Str.substr(0, 2)==L"0x")
        Result = wcstol(Str.c_str() + 2, NULL, 16);
    else
        Result = wcstol(Str.c_str(), NULL, 10);
    return (int)Result;
}
