//*#*#*#*#*#*#*#*#*#*#*#*#*#*#*#*#*#*#*#* DtmCmd.h *#*#*#*#*#*#*#*#*# (C) 2012-2019 DekTec
//
// DtmCmd - Command line tool for communication with DTM-32xx devices
//
// DtmCmd is a cross-platform (Windows & Linux) command-line tool for simple control
// of DTM-32xx devices. The user can read and write device settings, e.g. the command
// "DtmCmd -r 1 5" reads and prints the value of setting 5 in settings category 1. The
// most advanced capability of DtmCmd is upgrading the firmware of a DTM-32xx device.
// DtmCmd also serves as a real-world example of using the DTM protocol handler routines
// contained in DtmHandler.c and DtmHandler.h
// 
// See DtmCmd -? for help on available commands.
//
// You may modify DtmCmd for your specific needs, for example:
// - Change the interface implementation
// - Change the command line parameters
//
// DtmCmd implements the serial interface for Windows and Linux and the I2C interface 
// for Windows only.
//  - The Windows implementation of the serial interface operations uses CreateFile,
//    ReadFile and WriteFile, whereas the Linux serial implementation uses open, read 
//    and write.
//  - The I2C interface for Windows is based on a Nano River Tech MiniBoard I2C master
//    device and its API.
//
// This version of DtmCmd supports the DTM-3200, DTM-3224 and DTM-3237.
//
// Changelog:
//  v1.0.1 - First version released to the field
//  v1.1.0 - Added Busy error code
//         - Bug fix: Print unsigned DTM settings as unsigned instead of signed
//         - Added firmware upgrade example for non-USB interfaces in showhelp DtmCmd -?
//         - Extended help information for -usb parameter
//         - Replaced libusb0.dll with an 32-bit version (it was 64-bit only)
//  v1.2.0 - Linux only change: Using tcsetattr() to initialize RS-232 instead of stty
//         - Linux only change: Using poll() instead of read() for time-out implementation
//  v1.2.1 - Touch up of comments and coding style
//  v1.3.0 - Support for IPv6 addresses added
//  v1.4.0 - Improved error handling in RxFunction()
//  v1.4.1 - Bug fix: 2-digit COM port
//  v1.5.0 - Enable output flow control for RS232, make use of v1.14.0 DtmHandler
//  v1.6.0 - Make use of v1.15.0 DtmHandler for new DTM-3237B cmds
//
//+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+ LICENSE +=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=

// Copyright (C) 2012-2023 DekTec Digital Video B.V.
//
// Redistribution and use in source and binary forms, with or without modification, are
// permitted provided that the following conditions are met:
//  1. Redistributions of source code must retain the above copyright notice, this list
//     of conditions and the following disclaimer.
//  2. Redistributions in binary format must reproduce the above copyright notice, this
//     list of conditions and the following disclaimer in the documentation.
//  3. The source code may not be modified for the express purpose of enabling hardware
//     features for which no genuine license has been obtained.
//
// THIS SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED,
// INCLUDING BUT NOT LIMITED TO WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
// PURPOSE AND NONINFRINGEMENT.
// IN NO EVENT SHALL DEKTEC DIGITAL VIDEO BV, ITS AGENTS OR ITS EMPLOYEES BE LIABLE FOR
// ANY DIRECT, INDIRECT, CONSEQUENTIAL, INCIDENTAL, OR OTHER DAMAGES (INCLUDING DAMAGES
// FOR THE LOSS OF USE, INFORMATION, GOODWILL, PROFIT, WORK STOPPAGE, DATA, BUSINESS OR
// REVENUE) UNDER ANY CIRCUMSTANCES, OR UNDER ANY LEGAL THEORY, WHETHER IN CONTRACT, IN
// TORT, IN NEGLIGENCE, OR OTHERWISE, ARISING FROM THE USE OF, OR INABILITY TO USE THIS
// SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGES.

//.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.- Include files -.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-

#ifdef WINBUILD
    #include <Windows.h>            // Specific Windows types such as DWORD, HANDLE
#else
    #include <stdlib.h>
    #include <string.h>
    #include <unistd.h>
    #include <limits.h>
    #include <fcntl.h>
    #include <sys/stat.h>
    #include <errno.h>
    #include <stdio.h>
    #include <math.h>
    #include <poll.h>
    #include <termios.h>
#endif

// Windows/Microsoft uses with _ => Linux uses without
#ifndef WINBUILD
    #define _vsnprintf vsnprintf
    #define _vsnwprintf vswprintf
#endif

// Include files for I2C MiniBoard from Nano River Technologies
#ifdef WINBUILD
    #include "Miniboard.h"          // MiniBoard types and constants
    #include "MiniboardClass.h"     // MiniBoard control functions
    #include "Usb.h"                // USB definitions
#endif
#include "DtOpt.h"                  // DekTec command line option handler
#include "DtmHandler.h"             // Communication handler for DTM-32xx devices
#include <vector>                   // C++ standard vector implementation

//.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.- DtmCmd Version -.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.

// Version consists of three numbers:            // Changes when:
#define DTMCMD_VERSION_MAJOR        1            //  Major function(s) added or changed
#define DTMCMD_VERSION_MINOR        6            //  Minor function(s) added or changed
#define DTMCMD_VERSION_BUGFIX       0            //  Bugs are fixed, no change in function

//.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.- Constants -.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-

// Filename of DtmCmd's configuration file
const char  DTMCMD_CONFIG_FILE[] =  "DtmCmd.ini";

#define DTMCMDINI_BUFFER            1024         // Length of buffer to read ini file
#define MAX_BAUDRATE                256000       // Maximum supported baud rate
#define MAX_DTMADDRESS              255          // Maximum DTM address

//.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.- DtmCmdLine -.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.
//
// Decoded command-line arguments
//
struct DtmCmdLine
{
    int  m_Category;                // Category of DTM-32xx setting
    union
    {
        long long  m_Data;          // Data to be written
        unsigned char  m_Bytes[16];  // Data to be written >8bytes
    };
    wstring  m_FwUpgrFile;          // Filename of firmware upgrade file
    int  m_Index;                   // Index number
    int  m_Setting;                 // Setting number of DTM-32xx setting

    // The following arguments can be overwritten by entries from DtmCmd.ini
    wstring  m_Interface;           // Interface from command line or DtmCmd.ini
    wstring  m_SerialPath;          // Serial path from command line or DtmCmd.ini
    wstring  m_SerialBaud;          // Baudrate of serial control connection
    wstring  m_DtmAddress;          // DTM address from command line or DtmCmd.ini

    // Derived parameters from command line arguments
    int  m_CategoryIdx;             // Index of category in g_DtmCategoryInfo
    DtmDataTypes  m_DataType;       // Data type of setting
    int  m_SettingsIdx;             // Index of setting in g_DtmSettingInfo
    int  m_Units;                   // 10 => Setting is specified in units of 0.1
};

//.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.- DtmCmd -.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.
//
// Application class for DtmCmd.
// Everything is public (struct instead of class) so that the callback functions can
// access anything with ease.
//
struct DtmCmd
{
    DtmCmdLine  m_CmdLine;          // Decoded command-line arguments
    FILE*  m_pFwUpgrFile;           // File pointer to firmware upgrade file

    // Serial port handle differs between Windows/Linux
#if defined(WIN32) || defined(WIN64)
    HANDLE  m_hSerialPort;          // Handle to COM port connected to DTM-32xx
    MBc  m_I2cMiniBoard;            // Handle to I2C MiniBoard
#else
    int  m_SerialPort;              // Handle to COM port connected to DTM-32xx
#endif

    // Implemantation functions
    void  CloseSerialInterface();
    void  CreateDtmCmdIniFile(wstring Settings);
    void  DisplayInfo();
    void  InitI2cInterface();
    void  InitSerialInterface();
    void  LoadDtmCmdIni();
    int   Main(int Argc, char* Argv[]);
    void  ParseCommandLine(int argc, char* argv[]);
    void  PrintfDtmError(DtmResult Result);
    void  SendDtmCommand();
    void  ShowHelp();
    void  StoreInterfaceSettings();
};

//-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.- Exc -.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.
//
// Exception class
//
class Exc
{
public:
    Exc( const char* pFormat, ...  )
    {
        va_list  ArgList;
        va_start(ArgList, pFormat);
        _vsnprintf(m_ErrorMsg, sizeof(m_ErrorMsg)-1, (char*)(pFormat), ArgList);
        va_end(ArgList);
    };
    virtual ~Exc() {};

    operator const char* () const { return m_ErrorMsg; }

protected:
    // Error message
    char  m_ErrorMsg[1024];
};

//-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.- Options -.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.
//
// Structure to store command-line options
//
struct Options
{
    DtOpt  m_BaudRate;              // -baudrate: Set serial port rate (e.g. 9600)
    DtOpt  m_DtmAddress;            // -addr: Specify DTM address (e.g. 0x40)
    DtOpt  m_FirmwareUpgrade;       // -upgrade: Upgrade firmware
    DtOpt  m_Index;                 // -i: Index number
    DtOpt  m_Interface;             // -interface: Set type of interface
    DtOpt  m_Read;                  // -r: Read command
    DtOpt  m_Serial;                // -serial: Set path of the serial port (linux)
                                    //          or name of the com port (windows)
    DtOpt  m_Verbose;               // -v: Verbose mode
    DtOpt  m_ShowHelp;              // -?: Show help
    DtOpt  m_UsbIf;                 // -usb: Use USB interface
    DtOpt  m_Write;                 // -w: Write command
};

//.-.-.-.-.-.-.-.-.-.-.-.-.-.- DTM Handler Callback Functions -.-.-.-.-.-.-.-.-.-.-.-.-.-.

DtmResult  RxFunction(void* pOpaque, int DtmAddr, char* pData,
                                               int Expected, int* pReturned, int TimeOut);
DtmResult  SleepFunction(void* pOpaque, int Milliseconds);
DtmResult  TxFunction(void* pOpaque, int DtmAddr, char* pData, int Length);
DtmResult  UpgGetData(void* pOpaque, char* pData, int NumToRead);
void  UpgReportProgress(void* pOpaque, DtmUpgPhase Phase, int Progress);
void  PumpGuiFunction(void* pOpaque);

//-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.- Helper Functions -.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-

void  Log(const wchar_t* pMessage);
void  Log(const char* pMessage);
void  LogF(const wchar_t* pMessage, ... );
void  Sprintf(char* pDest, int Size, const char* pFormat, ...);
void  VSprintf(char* pDest, int Size, const char* pFormat, va_list VaArgs);
void  WsToCs(char* pDest, const wchar_t* pSource, int Size);
int  wtoi(wstring Str);
